#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from ZODB import DB
from ZEO import ClientStorage

from zeo_wrapper_prototype import ZEOWrapperPrototype


# Functions & classes =========================================================
class ZEOWrapper(ZEOWrapperPrototype):
    """
    ZEO wrapper based on ZEO client XML configuration file.

    Attributes:
        conf_path (str): Path to the configuration file.
        project_key (str): Project key, under which will this object access the
            ZEO structure.
        default_type (obj): Default data object used for root, if the root
            wasn't already created in ZEO.
    """
    def __init__(self, server, port, project_key=None,
                 run_asyncore_thread=True):
        """
        Initialize the object.

        Args:
            conf_path (str): See :attr:`conf_path`.
            project_key (str, default None): See :attr:`project_key`. If not
                set, the root of the database is used (this may cause
                performace issues).
            run_asyncore_thread (bool, default True): Run external asyncore
                thread, which handles connections to database? Default True.
        """
        self.server = server
        self.port = port

        super(ZEOWrapper, self).__init__(
            project_key=project_key,
            run_asyncore_thread=run_asyncore_thread,
        )

    def _get_db(self):
        """
        Open the connection to the database based on the configuration file.
        """
        return DB(
            ClientStorage.ClientStorage((self.server, self.port))
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import transaction


# Functions & classes =========================================================
def transaction_manager(fn):
    """
    Decorator which wraps whole function into ``with transaction.manager:``.
    """
    def transaction_manager_decorator(*args, **kwargs):
        with transaction.manager:
            return fn(*args, **kwargs)

    return transaction_manager_decorator



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import pytest

import transaction

from zeo_connector import ZEOWrapper
from zeo_connector import ZEOConfWrapper

from zeo_connector_defaults import generate_environment
from zeo_connector_defaults import cleanup_environment
from zeo_connector_defaults import tmp_context_name


# Variables ===================================================================
PROJECT_KEY = "some_key"


# Setup =======================================================================
def setup_module(module):
    generate_environment()


def teardown_module(module):
    cleanup_environment()


# Fixtures ====================================================================
@pytest.fixture
def zeo_conf_wrapper():
    return ZEOConfWrapper(
        conf_path=tmp_context_name("zeo_client.conf"),
        project_key=PROJECT_KEY,
    )


@pytest.fixture
def zeo_wrapper():
    return ZEOWrapper(
        server="localhost",
        port=60985,
        project_key=PROJECT_KEY,
    )


# Tests =======================================================================
def test_zeo_conf_wrapper_storing_and_retreiving():
    first_wrapper = zeo_conf_wrapper()
    second_wrapper = zeo_conf_wrapper()

    with transaction.manager:
        first_wrapper["something"] = "hello"

    with transaction.manager:
        assert second_wrapper["something"] == "hello"


def test_zeo_conf_wrapper_storing(zeo_conf_wrapper):
    with transaction.manager:
        zeo_conf_wrapper["azgabash"] = "hello"


def test_zeo_conf_wrapper_retreiving(zeo_conf_wrapper):
    with transaction.manager:
        assert zeo_conf_wrapper["azgabash"] == "hello"


def test_zeo_wrapper_retreiving(zeo_wrapper):
    with transaction.manager:
        assert zeo_wrapper["azgabash"] == "hello"


def test_zeo_wrapper_storing(zeo_wrapper):
    with transaction.manager:
        zeo_wrapper["zeo"] = "hello ZEO"


def test_zeo_wrapper_retreiving_again(zeo_wrapper):
    with transaction.manager:
        assert zeo_wrapper["zeo"] == "hello ZEO"


def test_dict_methods(zeo_wrapper, zeo_conf_wrapper):
    with transaction.manager:
        zeo_wrapper["first"] = 1

    used = {
        "something": "hello",
        "azgabash": "hello",
        "zeo": "hello ZEO",
        "first": 1,
    }

    with transaction.manager:
        assert "first" in zeo_conf_wrapper
        assert zeo_conf_wrapper.get("first", None) == 1
        assert zeo_conf_wrapper.get("second", 2) == 2

        assert set(zeo_conf_wrapper.keys()) == set(used.keys())
        assert set(zeo_conf_wrapper.values()) == set(used.values())

        assert set(zeo_conf_wrapper.iterkeys()) == set(used.iterkeys())
        assert set(zeo_conf_wrapper.itervalues()) == set(used.itervalues())

        iterated = {
            key: val
            for key, val in zeo_conf_wrapper
        }

        assert iterated == used

    with transaction.manager:
        del zeo_conf_wrapper["first"]

    with transaction.manager:
        assert "first" not in zeo_wrapper


def test_root_access():
    root = ZEOConfWrapper(
        conf_path=tmp_context_name("zeo_client.conf"),
    )

    assert PROJECT_KEY in root
    assert list(root.keys()) == [PROJECT_KEY]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from zeo_wrapper import ZEOWrapper
from zeo_conf_wrapper import ZEOConfWrapper
from zeo_wrapper_prototype import ZEOWrapperPrototype
from transaction_manager import transaction_manager



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import thread
import asyncore

import transaction
from ZODB.POSException import ConnectionStateError
from BTrees.OOBTree import OOBTree


# Variables ===================================================================
_ASYNCORE_RUNNING = False


# Functions & classes =========================================================
def _init_zeo():
    """
    Start asyncore thread.
    """
    if not _ASYNCORE_RUNNING:
        def _run_asyncore_loop():
            asyncore.loop()

        thread.start_new_thread(_run_asyncore_loop, ())

        global _ASYNCORE_RUNNING
        _ASYNCORE_RUNNING = True


def retry_and_reset(fn):
    """
    Decorator used to make sure, that operation on ZEO object will be retried,
    if there is ``ConnectionStateError`` exception.
    """
    def retry_and_reset_decorator(*args, **kwargs):
        obj = kwargs.get("self", None)

        if not obj:
            obj = args[0]

        try:
            return fn(*args, **kwargs)
        except ConnectionStateError:
            obj._on_close_callback()

        return fn(*args, **kwargs)

    return retry_and_reset_decorator


class ZEOWrapperPrototype(object):
    """
    ZEO wrapper prototype baseclass.

    Attributes:
        project_key (str): Project key, under which will this object access the
            ZEO structure.
        default_type (obj): Default data object used for root, if the root
            wasn't already created in ZEO.
    """
    def __init__(self, project_key=None, run_asyncore_thread=True):
        """
        Initialize the object.

        Args:
            conf_path (str): See :attr:`conf_path`.
            project_key (str, default None): See :attr:`project_key`. If not
                set, the root of the database is used (this may cause
                performace issues).
            run_asyncore_thread (bool, default True): Run external asyncore
                thread, which handles connections to database? Default True.
        """
        self.project_key = project_key
        self.default_type = OOBTree

        self._root = None  #: Reference to the root of the database.
        self._connection = None  #: Internal handler for the ZEO connection.

        if run_asyncore_thread:
            _init_zeo()

        self._open_connection()
        self._init_zeo_root()

    def _on_close_callback(self):
        """
        When the connection is closed, open it again and get new reference to
        ZEO root.
        """
        self._open_connection()
        self._init_zeo_root()

    def _get_db(self):
        """
        This should return the ZODB database object.
        """
        raise NotImplementedError("._get_db() is not implemented yet.")

    def _open_connection(self):
        """
        Open the connection to the database based on the configuration file.
        """
        db = self._get_db()
        self._connection = db.open()

        self._connection.onCloseCallback(self._on_close_callback)

    def _init_zeo_root(self, attempts=3):
        """
        Get and initialize the ZEO root object.

        Args:
            attempts (int, default 3): How many times to try, if the connection
                was lost.
        """
        try:
            db_root = self._connection.root()
        except ConnectionStateError:
            if attempts <= 0:
                raise

            self._open_connection()
            return self._init_zeo_root(attempts=attempts-1)

        # init the root, if it wasn't already declared
        if self.project_key and self.project_key not in db_root:
            with transaction.manager:
                db_root[self.project_key] = self.default_type()

        self._root = db_root[self.project_key] if self.project_key else db_root

    def sync(self):
        """
        Sync the connection.

        Warning:
            Don't use this method, if you are in the middle of transaction, or
            the transaction will be aborted.

        Note:
            You don't have to use this when you set :attr:`run_asyncore_thread`
            to ``True``.
        """
        self._connection.sync()

    @retry_and_reset
    def __getitem__(self, key):
        return self._root[key]

    @retry_and_reset
    def __setitem__(self, key, val):
        self._root[key] = val

    @retry_and_reset
    def __contains__(self, key):
        return key in self._root

    @retry_and_reset
    def __delitem__(self, key):
        del self._root[key]

    @retry_and_reset
    def __iter__(self):
        return self._root.iteritems()

    @retry_and_reset
    def iteritems(self):
        return self._root.iteritems()

    @retry_and_reset
    def keys(self):
        return self._root.keys()

    @retry_and_reset
    def iterkeys(self):
        return self._root.iterkeys()

    @retry_and_reset
    def values(self):
        return self._root.values()

    @retry_and_reset
    def itervalues(self):
        return self._root.itervalues()

    @retry_and_reset
    def get(self, key, alt):
        return self._root.get(key, alt)

    def pack(self):
        """
        Call .pack() on the database (transaction history cleanup).
        """
        return self._get_db().pack()
